#include <err.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <sys/io.h>

#include "rubyclk.h"

int main(int argc, char *argv[])
{
	static const unsigned char req[5] = {3, 5, 0x10, 0x19, 0x34};

	unsigned int data;
	unsigned char got[5];

	unsigned char sysclk = 0; /* default */
	if(argc > 1)
		sysclk = rubyclk(argv);
	/*
	 * grant access to ports
	 *
	 * 0xcf8 (32bit) and 0xcfc (32bit), 8 bytes in total
	 * 0x4e (8bit) and 0x4f (8bit), 2 bytes in total
	 */
	if(ioperm(0xcf8, 8, true)
	|| ioperm(0x4e, 2, true))
		err(EXIT_FAILURE, "ioperm");
	/*
	 * PCI configuration space:
	 *
	 * 0xcf8: configuration space address (0x80000000 | bus << 16 | device << 11 | function << 8 | offset)
	 * 0xcfc: configuration space data
	 */
	outl(0x8000f800, 0xcf8);
	data = inl(0xcfc);

	/*
	 * vendor ID in lower 16bit, device ID in upper 16bit
	 *
	 * vendor 8086: Intel
	 * device 27b8: 82801GB/GR (ICH7 Family) LPC Interface Bridge
	 */
	if(data != 0x27b88086) {
		/* close PCI config */
		outl(0, 0xcf8);
		errx(EXIT_FAILURE, "LPC controller on Intel ICH7 not found! Got 0x%08x instead of 0x27b88086.", data);
	}

	/* write 0x26 twice to enable F85226FG configuration registers */
	outb(0x26, 0x4e);
	outb(0x26, 0x4e);
	/* chip ID 1 */
	outb(0x5a, 0x4e);
	got[0] = inb(0x4f);
	/* chip ID 2 */
	outb(0x5b, 0x4e);
	got[1] = inb(0x4f);
	/* version */
	outb(0x5c, 0x4e);
	got[2] = inb(0x4f);
	/* vendor ID 1 */
	outb(0x5d, 0x4e);
	got[3] = inb(0x4f);
	/* vendor ID 2 */
	outb(0x5e, 0x4e);
	got[4] = inb(0x4f);

	/* check power-on defaults */
	if(memcmp(got, req, 5)) {
		/* close PCI config */
		outl(0, 0xcf8);
		errx(EXIT_FAILURE, "Fintek F85226FG not found! Got 0x%02x-%02x-%02x-%02x-%02x instead of 0x03-05-10-19-34.",
				(unsigned)got[0], (unsigned)got[1], (unsigned)got[2], (unsigned)got[3], (unsigned)got[4]);
	}

	/* 200-2ff */
	outl(0x8000f884, 0xcf8);
	outb(0x01, 0xcfc);
	outb(0x02, 0xcfd);
	outb(0xfc, 0xcfe);
	/* 380-38f, 390-39f */
	outl(0x8000f888, 0xcf8);
	outb(0x89, 0xcfc);
	outb(0x03, 0xcfd);
	outb(0x1c, 0xcfe);
	/* a00-aff */
	outl(0x8000f88c, 0xcf8);
	outb(0x01, 0xcfc);
	outb(0x0a, 0xcfd);
	outb(0xfc, 0xcfe);
	/* 300/310/320/330/340/350/360/370+3 */
	outl(0x8000f890, 0xcf8);
	outb(0x01, 0xcfc);
	outb(0x03, 0xcfd);
	outb(0x70, 0xcfe);

	/* close PCI config */
	outl(0, 0xcf8);

	/* 200-2ff */
	outb(0x21, 0x4e);
	outb(0x00, 0x4f);
	outb(0x22, 0x4e);
	outb(0x02, 0x4f);
	outb(0x20, 0x4e);
	outb(0xff, 0x4f);
	/* 380-38f, 390-39f */
	outb(0x24, 0x4e);
	outb(0x88, 0x4f);
	outb(0x25, 0x4e);
	outb(0x03, 0x4f);
	outb(0x23, 0x4e);
	outb(0x1f, 0x4f);
	/* a00-aff */
	outb(0x31, 0x4e);
	outb(0x00, 0x4f);
	outb(0x32, 0x4e);
	outb(0x0a, 0x4f);
	outb(0x30, 0x4e);
	outb(0xff, 0x4f);
	/* 300/310/320/330/340/350/360/370+3 */
	outb(0x34, 0x4e);
	outb(0x00, 0x4f);
	outb(0x35, 0x4e);
	outb(0x03, 0x4f);
	outb(0x33, 0x4e);
	outb(0x73, 0x4f);
	/* enable A19, A18, and A17 */
	outb(0x05, 0x4e);
	outb(0x0e, 0x4f);
	/* ISA speed and waitstates */
	outb(0x06, 0x4e);
	outb(sysclk, 0x4f);

	/* enable SYSCLK */
	outb(0x50, 0x4e);
	outb(0x00, 0x4f);
	/* no wakeup/sleep on PWRDN, and IRQ[1,3-7] */
	outb(0x51, 0x4e);
	outb(0x00, 0x4f);
	/* no wakeup/sleep on IRQ[8-12,14,15] */
	outb(0x52, 0x4e);
	outb(0x00, 0x4f);

	return EXIT_SUCCESS;
}
